<?php
namespace Models;

// Модель для работы с дисплеем покупателя
class Cd extends \DB\SQL\Mapper{
	public $db, $f3, $web, $sets, $method='POST', $messages;

	public function __construct($db, $f3) {
		parent::__construct($db, 'cd_settings' ); // подключаемся к таблице с настройками
		$this->f3=$f3;
		$this->db=$db;
		$this->web=\Web::instance();
		$this->messages=new \Models\Messages();
		$this->sets=$this->get_cd();
	}

	// Узнаем настройки дисплея покупателя
	function get_cd() {
		$this->reset();
		$this->load();
		$sets=$this->cast();
		return $sets;
	}

	// Сохраняем настройки дисплея покупателя
	function save_cd($post) {
		$this->reset();
		$this->get_cd();
		$this->copyFrom($post);
		$this->save();
		if($this->dry()) {
			return array(
				'success'=>false,
				'txt'=>'db_error'
			);
		}
		return array(
			'success'=>true,
			'txt'=>'save_successful'
		);
	}

	// Тест сервера
	function test($driver, $restart=true) {
		// Тестовый запрос
		$url=$this->sets['url'];
		$options = array(
			'timeout'=>3,
    		'method'  => $this->method,
    		'content' => "data=test;\n"
		);
		$res=$this->web->request($url, $options);
		// Если в ответ не пришел код ответа 200, то надо запустить сервер
		if(mb_strpos($res['headers'][0], '200')===false) {
			if($driver!='') {
				if($restart) {
					$script=$driver.'/server.py';
					include('drivers/start.php');
					$this->test($driver, false);
				}
				else {
					echo "Не удается запустить драйвер дисплея покупателя";
					return;
				}
			}
			else {
				echo "Сервер дисплея покупателя не доступен";
				return;
			}
		}
	}

	// Печать на дисплее покупателя
	function print($txt, $timeout=3) {
		if(!$this->sets['active']) return;
		$this->test($this->sets['driver']);
		$url=$this->sets['url'];
		$options = array(
			'timeout'=>$timeout,
    		'method'  => $this->method,
    		'content' => 'data='.$txt
		);
		$res=$this->web->request($url, $options);
	}

	// Очистить дисплей покупателя
	function clear_cd() {
		$this->print("p;;;");
	}

	// Текст закрытой кассы
	function close_txt() {
		$this->print($this->sets['close_txt']);
	}

	// Текст готовности к новой продаже
	function seller_txt() {
		$this->print($this->sets['seller_txt']);
	}

	// Текст добавления товара в чек
	function print_line($line, $count) {
		$price=$line['price']+($line['price']*($line['vat']/100))*max(0, $line['vat_type']);
		$txt="p;".$line['product_title'].";";
		$txt.=$count."x".number_format($price, 2, '.', '')."=".number_format($price*$count, 2, '.', '').";";
		$this->print($txt);
	}

	// Текст при оплате чека
	function total_price($sum, $pay) {
		$txt="p;Итого:".number_format($sum, 2, '.', '').";";
		if($pay-$sum>0) {
			$txt.="Сдача:".number_format($pay-$sum, 2, '.', '');
		}
		$txt.=";";
		$this->print($txt);
	}

	// Список драйверов
	function get_drivers() {
		$drivers=array();
		$displays_dir=$this->f3->get('displays_dir');
		$dirs=scandir($displays_dir); // Ищем все драйвера
		if($dirs) {
			foreach ($dirs as $d) {
				$folder=$displays_dir.$d;
				$file=$folder.'/server.config';
				if(is_dir($folder) && file_exists($file)) {
					$config=parse_ini_file($file);
					if($config) {
						$drivers[$folder]=array(
							'title'=>$config['ServerName'],
							'path'=>$folder
						);
					}
				}
			}
		}

		return $drivers;
	}
}
