<?php
namespace Models;

// Модель Залов
class Rooms extends \DB\SQL\Mapper {
	public $db, $f3, $tables;

	public function __construct($db, $f3) {
		parent::__construct($db, 'rooms' ); // подключаемся к таблице
		$this->db=$db;
		$this->f3=$f3;
		$this->tables=new \Models\Tables($db, $f3); // модель столов
	}

	// Информация об открываемой странице
	function page_data($page='sales') {
		$pages=array(
			'add'=>array(
				'title'=>'Добавить зал',
				'inc'=>'rooms_form.htm'
			),
			'change'=>array(
				'title'=>'Редактировать зал',
				'inc'=>'rooms_form.htm'
			),
			'select_table'=>array(
				'title'=>'Продажа',
				'inc'=>'rooms_select_table.htm'
			)
		);
		$pages[$page]['page']=$page;
		return $pages[$page];
	}

	// Выбор из базы
	function select_rooms($query=null, $options=null) {
		$this->reset();
		$this->load($query, $options);
		$rooms=array();
		if(!$this->dry()) {
			do {
				$rooms[]=$this->cast();
			}
			while ($this->skip());
		}
		return $rooms;
	}

	// Сохраняем зал
	function save_room($data) {
		if(empty($data['title'])) {
			return array(
				'success'=>false,
				'txt'=>'not_enough_data'
			);
		}
		$this->reset();
		$this->load(array('title=? AND id!=?', array($data['title'], $data['id'])), array('limit'=>1));
		if(!$this->dry()) {
			return array(
				'success'=>false,
				'txt'=>'room_is_exists'
			);
		}
		$this->reset();
		if($data['id']) {
			$this->load(array('id=?', $data['id']), array('limit'=>1));
		}
		unset($data['id']);
		$this->copyFrom($data);
		if($this->save()!==false) {
			return array(
				'success'=>true,
				'txt'=>'save_successful',
				'reload'=>'.rooms_list',
				'close_all_window'=>true,
				'function'=>'open_room("/settings/rooms?id='.$this->get('_id').'")'
			);
		}
		else {
			return array(
				'success'=>false,
				'txt'=>'db_error'
			);
		}
	}

	// Удаление зала
	function delete($id) {
		if($id>0) {
			// Проверяем наличие зала
			$this->reset();
			$this->load(array('id=?',$id), array('limit'=>1));
			if($this->dry()) {
				return array(
					'success'=>false,
					'txt'=>'not_found'
				);
			}
			// Проверяем наличие столов, их не должно быть
			$this->tables->reset();
			$this->tables->load(array('room=?', $id));
			if(!$this->tables->dry()) {
				return array(
					'success'=>false,
					'txt'=>'room_is_not_empty'
				);
			}
			if($this->erase()) {
				return array(
					'success'=>true,
					'txt'=>'delete_successful',
					'reload'=>'.rooms_list',
					'close_all_window'=>true,
					'function'=>'open_room("/settings/rooms")'
				);
			}
			else {
				return array(
					'success'=>false,
					'txt'=>'db_error'
				);
			}
		}
		else {
			return array(
				'success'=>false,
				'txt'=>'not_enough_data'
			);
		}
	}
}
