<?php
/**
 * для штрих алгоритм такой:
 * - отправляем пакет с командой: STX, <LEN>, <PASS>, <MSG>, <CRC>
 * - ждем ответа ФР, должен быть ACK
 * - пауза, возможно очень маленькая, возможно большая, пока ФР готовит ответ
 * - читаем ответ ФР
 * - отвечаем ACK
 * 
 * формат запроса, например гудок: 0x13
 * STX, 05, 13, 1E 00 00 00, 08
 * формат ответа такой же:
 * ACK, STX, 03 13 00 1E 0E
 * 
 * Если в процессе печати документа произошёл обрыв ленты, то на ней, печатается строчка
«**ОБРЫВ БУМАГИ ДОКУМЕНТ НЕЗАВЕРШЕН**» и печать приостанавливается. АСПД
переходит в подрежим 2 «Активное отсутствие бумаги». Оператору требуется установить
новый рулон в АСПД согласно инструкции по заправке бумаги (см. соответствующий раздел
выше). При этом АСПД переходит в подрежим 3 «После активного отсутствия бумаги». 
Затем оператор должен подать команду B0h «Продолжение печати» (все другие команды, 
связанные с печатью, блокируются в подрежиме 3). После подачи команды продолжения
печати прерванный документ повторяется
 * 
 * 25h - отрезка чека
 * 
 * @author  mexxval
 * @link    http://blog.sci-smart.ru
 */

class ShtrihM_HiLevel extends KKM {
    // номера битов в битовых полях состояния ККМ
    const FLAGS2_SESSION_OPENED = 6; // 0 - closed, 1 - opened
    const FLAGS2_24HOUR_PASS = 7; // 0 - no, 1 - yes
    const FLAGS_MONEYBOX_OPENED = 11; // 0 - closed, 1 - opened
    const FLAGS_CASE_OPENED = 10; // 0 - опущена, 1 - поднята
    const FLAGS_PAPER_OUT = 7; // 0 - yes, 1 - no
    const PAPER_WIDTH = 32; // Ширина бумаги

    public function isReady () {
        $ans = $this->enq();
        if ($this->isNAK($ans)) {
            return true;
        } else if ($this->isACK($ans)) {
            // это может понадобиться, чтобы прочитать непереданных от ФР пакет.
            // такое может быть если наш ACK не успел дойти до ФР или еще какая беда приключилась со связью
            $this->readAnswer(); // читаем, освобождаем буфер ФР
            return true;
        }
        return false;
    }

    protected function init() {
        // оператор1 - пароль 1 
        // оператор2 - пароль 2 [...]
        // оператор28 - пароль 28 
        // администратор - пароль 29 
        // системный администратор - пароль 30
        $this->setCashierPassword(30);
        parent::init();
    }

    // Формирование tlv
    protected function make_tlv($tag, $value, $s=5) {
        $value = iconv('UTF-8', 'cp866', $value);
        if($s==5) {
            $l=mb_strlen($value);
            $value=$this->prepareText($value, $l);
        }
        elseif($s==1) {
            $l=1;
            $value=$this->prepareArg($value, 1);
        }
        elseif($s==2) {
            $l=2;
            $value=$this->prepareArg($value, 2);
        }
        elseif($s==3) {
            $l=4;
            $value=$this->prepareArg($value, 4);
        }
        else {
            $l=mb_strlen($value)/2;
        }
        $tlv=$this->prepareArg($tag, 2).$this->prepareArg($l, 2).$value;
        return $tlv;
    }

    // Обработка команд согласно протоколу Qkkm
    public function wtf($s) {
        $s=trim($s);
        $p=explode(';', $s);
        $n=count($p);
        if($p[$n-1]=='') {
            unset($p[$n-1]);
            $n-=1;
        }
        if($p[0]=='feed') {
            $res.=$this->feed($p[1]);
        }
        elseif($p[0]=='open_session') { // Открытие смены
            $res=$this->openSession();
        }
        elseif($p[0]=='p' || $p[0]=='pm') { // Печать текста
            $res=$this->p($p[1]);
        }
        elseif($p[0]=='print_font') { // Печать строки заданным шрифтом
            $res=$this->print_font($p[1], $p[2]);
        }
        elseif($p[0]=='print_bold') { // Печать жирной строки
            $res=$this->print_bold($p[1]);
        }
        elseif($p[0]=='x') { // X отчет
            $res=$this->x();
        }
        elseif($p[0]=='z') { // Z отчет
            $res=$this->z();
        }
        elseif($p[0]=='beep') { // гудок
            $res=$this->beep();
        }
        elseif($p[0]=='continue_print') { // Продолжить печать чека
            $res=$this->continue_print();
        }
        elseif($p[0]=='c') { // Печать копии последнего чека
            $res=$this->c();
        }
        elseif($p[0]=='g') { // Обнуление чека
            $res=$this->cancelCheck();
        }
        elseif($p[0]=='d') { // Статус ККМ
            $res=$this->d();
        }
        elseif($p[0]=='imde') { // Внесенеи или изъятие из кассы
            $res=$this->imde($p[1]);
        }
        elseif($p[0]=='set_tlv') { // Установка парметра TLV
            $res=$this->set_tlv($p[1], $p[3], $p[2]);
        }
        elseif($p[0]=='b') { // Открытие чека
            $res=$this->b($p[1], $p[2], $p[3]);
        }
        elseif($p[0]=='smde') { // Добавление товара в чек
            $res=$this->smde($p[1], $p[2], $p[3], $p[4], $p[5], $p[6], $p[7]);
        }
        elseif($p[0]=='tmde') { // Оплата и закрытие чека
            if($n==3) {
                $res=$this->tmde_ffd_105(array($p[2]=>$p[1]), array($p[2]=>0)); // ффд 1.05
            }
            elseif($n==9) {
                $res=$this->tmde_ffd_105(array($p[1], $p[2], $p[3], $p[4]), array($p[5], $p[6], $p[7], $p[8])); // ффд 1.05
            }
            elseif($n==7) {
                $res=$this->tmde_ffd_11(array($p[1], $p[2], $p[3], $p[4], $p[5]), $p[6]); // ФФД 1.1
            }
            else {
                $res=array(false, 'txt'=>$this->getErrorMessagebyCode(9));
            }
        }
        elseif($p[0]=='cut_check') { // Отрезка чека
            $res=$this->cutCheck(1);
        }
        elseif($p[0]=='m') { // Запрос денежного регистра
            $res=$this->m($p[1]);
        }
        elseif($p[0]=='open_cash_box') {
            $res=$this->open_cash_box($p[1]); // Открытие денежного ящика
        }
        elseif($p[0]=='correction') { // Чек коррекции
            $res=$this->correction($p[1], $p[2]);
        }
        elseif($p[0]=='print_clishe') {
            $res=$this->print_clishe();
        }
        elseif($p[0]=='print_header') {
            $res=$this->print_header($p[1], $p[2]);
        }
        elseif($p[0]=='cmd') { // Произвольная команда
            $res=$this->cmd($p);
        }
        elseif($p[0]) {
            $res=array(false, 'txt'=>$this->getErrorMessagebyCode(1));
        }
        elseif($p[0]=='') {
            $res='ack';
        }
        return $res;
    }

    // Ожидание готовности принтера
    public function wait() {
        $c=100;
        while ($c>0) {
            $c--;
            if($this->sets['WaitTime']>0) {
                usleep($this->sets['WaitTime']);
            }
            if(!$this->isPrintInProgress()) {
                break;
            }
        }
    }

    // Форматирование текста
    function text_format($s, $w=self::PAPER_WIDTH) {
        $s=str_replace('#kkm_bold#', '', $s);
        // Заменяем #kkm_right# на пробелы
        $kkm_right="#kkm_right#"; # тэг сдвига вправо
        $space_n=$w-(mb_strlen($s)-mb_strlen($kkm_right)); // кол-во пробелов
        $s=str_replace($kkm_right, str_repeat(" ", $space_n), $s); // заменяем тэг сдвига на пробелы

        // Заменяем #kkm_center# на пробелы
        $kkm_center="#kkm_center#"; // тэг сдвига в центр
        $space_n=floor(($w-(mb_strlen($s)-mb_strlen($kkm_center)))/2); // кол-во пробелов
        $s=str_replace($kkm_center, str_repeat(" ", $space_n), $s); // заменяем тэг сдвига на пробелы
        return $s;
    }

    // Выполнение произвольной команды
    public function cmd($p) {
        $cmd=$p[1];
        $args='';
        $n=count($p);
        if($n>2) {
            for($i=2; $i<$n; $i+=2) {
                $l=$p[$i+1];
                if(is_numeric($p[$i])) {
                    $args.=$this->prepareArg($p[$i], $l);
                }
                else {
                    $args.=$this->prepareText($p[$i], $l);
                }
            }
        }
        return $this->writecmd($cmd, $args);
    }

    // Печать заголовка чека
    public function print_header($title, $num) {
        $args=$this->prepareText($title, 30);
        $args.=$this->prepareArg($num, 2);
        return $this->writecmd('18', $args);
    }

    // Печать клише
    public function print_clishe() {
        return $this->writecmd('52');
    }

    // Чек коррекции
    public function correction($type, $sum) {
        $args=$this->prepareArg($sum, 5);
        if($type=='docIn') {
            $args.=$this->prepareArg(1, 1);
        }
        elseif($type=='docOut') {
            $args.=$this->prepareArg(2, 1);
        }
        elseif($type=='docRetIn') {
            $args.=$this->prepareArg(3, 1);
        }
        elseif($type=='docRetOut') {
            $args.=$this->prepareArg(4, 2);
        }
        else {
            return array(false, 'txt'=>$this->getErrorMessagebyCode(9));
        }
        echo $args;
        return $this->writecmd('ff36', $args);
    }

    // Открытие денежного ящика
    public function open_cash_box($box) {
        $args=$this->prepareArg($box, 1);
        return $this->writecmd('28', $args);
    }

    // Запрос денежного регистра
    public function m($reg) {
        $res=$this->getKKMMoneyRegister($reg);
        $this->msg=$res;
        return 'ack';
    }

    // Оплата чека, ФФД 1.1
    public function tmde_ffd_11($pays, $sno) {
        $args='';
        // Оплата
        $args.=$this->prepareArg($pays[0], 5); // Налом
        $args.=$this->prepareArg($pays[1], 5); // Электронно
        $args.='00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'; // Еще способы, не используются
        $args.=$this->prepareArg($pays[2], 5); // Предоплата
        $args.=$this->prepareArg($pays[3], 5); // Постоплата
        $args.=$this->prepareArg($pays[4], 5); // Встречное предложение
        $args.='00'; // Округление до рубля, нет
        $args.='000000000000'; // Налоги. 0 - расчитывается системой
        $args.=$this->prepareArg($sno, 1); // Сно
        $args .= $this->prepareText('', 64);
        return $this->writecmd('ff45', $args);
    }

    // Оплата чека, ФФД 1.05
    public function tmde_ffd_105($pays, $taxes) {
        $args='';

        // Способы оплаты
        for($i=0; $i<4; $i++) {
            if($pays[$i]) {
                $args.=$this->prepareArg($pays[$i], 5);
            }
            else {
                $args.=$this->prepareArg(0, 5);
            }
        }
        $args.='0000'; // Скидка или надбавка в %

        // Налоги
        for($i=0; $i<4; $i++) {
            if($taxes[$i]) {
                $args.=$this->prepareArg($taxes[$i], 1);
            }
            else {
                $args.=$this->prepareArg(0, 1);
            }
        }
        $args .= $this->prepareText('', 40);
        return $this->writecmd('85', $args);
    }

    // Добавляем товар в чек
    public function smde($title, $price, $qty, $vt, $section, $psr=null, $ppr=null) {
        // Узнаем тип открытого докмента
        $status_kkm=$this->getKKMState();
        if($status_kkm['fr_mode']=='08') {
            if($psr && $ppr) {
                return $this->operation_v2(1, $price, $qty/1000, $section, $title, $vt, $psr, $ppr);
            }
            return $this->sell($price/100, $qty/1000, $section, $title, $vt);
        }
        elseif($status_kkm['fr_mode']=='28') {
            if($psr && $ppr) {
                return $this->operation_v2(2, $price, $qty/1000, $section, $title, $vt, $psr, $ppr);
            }
            return $this->return_sell($price/100, $qty/1000, $section, $title, $vt);
        }
        return array(false, 'txt'=>$this->getErrorMessagebyCode(2));
        
    }

    // Открыть чек
    public function b($type, $cashier, $print) {
        $types=array(0=>0, 1=>2);
        $r=$this->writecmd('8D', $this->prepareArg($types[$type], 1));
        if($r===false || $r=='nak') return $r;
        return $this->set_tlv('1021', $cashier, 5);
    }

    // Установка tlv
    public function set_tlv($tag, $value, $s) {
        $tlv=$this->make_tlv($tag, $value, $s);
        return $this->writecmd('ff0c', $tlv);
    }

    // Внесение денег в кассу 
    public function imde($summ) {
        if($summ>0) {
            return $this->writecmd('50', $this->prepareArg(abs($summ), 5));
        }
        else {
            return $this->writecmd('51', $this->prepareArg(abs($summ), 5));
        }
    }

    // Запрос статуса ККМ
    public function d() {
        $res=$this->getKKMState();
        if(is_array($res)) {
            $this->msg=$res;
            return 'ack';
        }
        else {
            return 'nak';
        }
    }

    // Продолжение печати
    public function continue_print() {
        return $this->writecmd('B0');
    }

    // Печать копии последнего чека
    public function c() {
        return $this->writecmd('8C');
    }

    // Протяжка ленты
    public function feed($n) {
        return $this->writecmd('29',  $this->prepareArg('1', 1).$this->prepareArg($n, 1));
    }

    // Печать простой строки
    public function p($text) {
        $strs=explode('#kkm_br#', $text);
        foreach ($strs as $s) {
            if(mb_substr($s, 0, mb_strlen('#kkm_bold#'))=='#kkm_bold#') {
                $res=$this->print_bold($s);
            }
            else {
                $s=$this->text_format($s);
                $res=$this->writecmd('17',  $this->prepareArg('1', 1).$this->prepareText($s, 40));
            }
            if($res===false || $res=='nak') {
                break;
            }
        }
        return $res;
    }

    // Печать строки заданным шрифтом
    public function print_font($font, $text) {
        $strs=explode('#kkm_br#', $text);
        foreach ($strs as $s) {
            $s=$this->text_format($s, floor(self::PAPER_WIDTH/$font));
            $res=$this->writecmd('2F',  $this->prepareArg('1', 1).$this->prepareArg($font, 1).$this->prepareText($s, floor(40/$font)));
            if($res===false || $res=='nak') {
                break;
            }
        }
        return $res;
    }

    // Печать жирной строки
    public function print_bold($text) {
        $strs=explode('#kkm_br#', $text);
        foreach ($strs as $s) {
            $s=$this->text_format($s, floor(self::PAPER_WIDTH/2));
            $res=$this->writecmd('12',  $this->prepareArg('1', 1).$this->prepareText($s, 20));
            if($res===false || $res=='nak') {
                break;
            }
        }
        return $res;
    }

    // x-отчет
    public function x() {
        $res=$this->writecmd('40');
        return $res;
    }

    // z-отчет
    public function z() {
        $res=$this->writecmd('41');
        return $res;
    }

    public function beep() {
        return $this->writecmd('13');
    }

    public function getKKMState() {
        $answer = $this->writecmd('11');
        if($answer==='nak') {
            return 'nak';
        }
        $scheme = array(
            'lastcmd' => 1,
            'errcode' => 1,
            'cashieridx' => 1,
            'fr_soft_ver' => 2,
            'fr_soft_build' => 2,
            'fr_soft_date' => 3, // dd-mm-yy
            'idxinroom' => 1,
            'docno' => 2,
            'fr_flags' => 2,
            'fr_mode' => 1,
            'fr_submode' => 1,
            'fr_port' => 1,
            'fp_soft_ver' => 2,
            'fp_soft_build' => 2,
            'fp_soft_date' => 3,
            'date' => 3,
            'time' => 3,
            'fp_flags' => 1,
            'plant_no' => 4,
            'last_smena_no' => 2,
            'fp_free_rec' => 2,
            'rereg' => 1,
            'rereg_left' => 1,
            'inn' => 6,
        );
        $curstate = 0;
        $result = array();
        foreach ($scheme as $var => $len) {
            $part = substr($answer, $curstate, $len);
            $result[$var] = $this->unpackBinaryString($part);
            $curstate += $len;
        }
        $result['fr_flags'] = sprintf("%08b", hexdec($result['fr_flags']));
        $result['fp_flags'] = sprintf("%08b", hexdec($result['fp_flags']));
        $result['fr_mode_text'] = $this->getModeTitleByCode($result['fr_mode']);
        $result['fr_submode_text'] = $this->getSubModeTitleByCode($result['fr_submode']);
        return $result;
    }
    public function is24HourPass() {
        return $this->getStateFlag(self::FLAGS2_24HOUR_PASS, 2) == 1;
    }
    public function isSessionOpened() {
        return $this->getStateFlag(self::FLAGS2_SESSION_OPENED, 2) == 1;
    }
    public function isMoneyBoxOpened() {
        return $this->getStateFlag(self::FLAGS_MONEYBOX_OPENED) == 1;
    }
    public function isPaperOut() {
        return $this->getStateFlag(self::FLAGS_PAPER_OUT) == 0;
    }
    public function isCaseOpened() {
        return $this->getStateFlag(self::FLAGS_CASE_OPENED) == 1;
    }
    public function openMoneyBox($num = 0) {
        return $this->writecmd('28', sprintf('%02d', $num));
    }
    public function setDateConfirm($ts = null) {
        return $this->setDate($ts, '23');
    }
    // печать отчета или чего еще
    public function isPrintInProgress() {
        $state = $this->getKKMState();
        // контролируем submode
        if($state=='nak') return 'nak';
        return $state['fr_submode'] == '05';
    }
    public function setDate($ts = null, $cmd = '22') {
        if ($ts == null or !date('Ymd', $ts)) {
            $ts = time();
        }
        if (!in_array($cmd, array('22', '23'))) {
            $cmd = '22';
        }
        $time = sprintf('%02x%02x%02x', date('d', $ts), date('m', $ts), date('y', $ts));
        return $this->writecmd($cmd, $time);
    }
    public function setTime($ts = null) {
        if ($ts == null or !date('His', $ts)) {
            $ts = time();
        }
        $time = sprintf('%02x%02x%02x', date('H', $ts), date('i', $ts), date('s', $ts));
        return $this->writecmd('21', $time);
    }
    # аннулирование всего чека
    public function cancelCheck() {
        return $this->writecmd('88');
    }
    public function cutCheck($cut = 1) {
        $cut = (bool) $cut; // 1 - неполная, 0 - полная
        return $this->writecmd('25', sprintf('%02d', $cut));
    }
    /*
     * Тип документа (1 байт): 0 – продажа; 
    1 – покупка; 
    2 – возврат продажи; 
    3 – возврат покупки
     */
    public function openNewCheck($type = 1) {
        return $this->writecmd('8D', sprintf('%02d', $type));
    }
    # открыть смену
    public function openSession() {
        return $this->writecmd('E0');
    }


    protected function getStateFlag($bit, $reg = 1) {
        if (!($state = $this->getKKMState())) {
            throw new Exception('Failed to get KKM state');
        }
        $reg = $reg == 2 ? 'fp_flags' : 'fr_flags';
        $tmp = strrev($state[$reg]);
        return $tmp{$bit};
    }
    protected function makeBinaryCmd($cmdcode, $data, $wPass = true) {
        $passwd = $wPass ? sprintf("%02x000000", $this->getCashierPassword()) : '';
        error_log($data);
        $cmd = pack('H*', $cmdcode . $passwd) . $this->escapeData(pack('H*', $data));
        $msg_len = pack('C', strlen($cmd));
        $cmdcrc = $this->makecrc($msg_len . $cmd);
        $cmd = pack('C*', self::STX) . $msg_len . $cmd . pack('C*', $cmdcrc);
        return $cmd;
    }
    protected function escapeData($data) {
        return $data;
    }
    public function writecmd($cmdcode, $data = '') {
        if($cmdcode!='11') {
            $this->wait();
        }
        $wPass = !in_array($cmdcode, array('16'));
        $cmd = $this->makeBinaryCmd($cmdcode, $data, $wPass);
        $answer = false;
        try {
            $ack = $this->sendBinary($cmd, 5e5, 1); // таймаут 500мс на ожидание ACK
            if (is_null($ack)) { // timeout?
                //throw new Exception('Response wait timeout. Link not ready');
                $answer=false;
            } else if ($this->isACK($ack)) {
                $answer = $this->readAnswer($cmdcode);
            } else {
                $answer='nak';
                $this->log('NOT ACK'); // что делать в этом случае?
                // Если в ответ на сообщение ФР получен NAK, сообщение не повторяется, ФР ждет уведомления ENQ для повторения ответа
            }
        }
        catch (Exception $e) {
            throw $e;
        }
        if($this->sets['PortType']=='remote') {
            $this->readAnswer();
        }
        return $answer;
    }
    protected function readAnswer($cmdcode=null) {
        // пингуем enq-ами, если отвечает ack-ами, значит готовит данные, ждем
        $cycles = 100;
        // читаем первый байт ответа
        $ans = $this->readBinaryAnswer(1, 5e4); // таймаут начала передачи от ФР. может быть большим
        do {
            if ($this->isSTX($ans)) {
                break;
            }
            else if ($this->isACK($ans)) { // пытаемся читать ответ
                $this->log('<B');
                $ans = $this->enq();
                continue;
            }
            else if (is_null($ans)) { // timeout
                $ans = $this->enq();
                continue;
            }
            else {
                $this->log('Expected STX in the beginning of answer. Got ' . $this->byteCode($ans));
                return false;
            }
        } while ($cycles-- > 0);

        $len_b = $this->readBinaryAnswer(1, 5e4);
        $len = $this->byteCode($len_b);
        $ans = $this->readBinaryAnswer($len + 1, 5e4);
        if ($this->checkCRC($len_b . $ans)) {
            $this->log('CRC MATCH');
            $this->ack();
            // проверим код ошибки
            $n=mb_strlen($cmdcode)/2;
            $errcode = $this->byteCode($ans{$n}); // длина пропущена, после нее код команды, а затем байт ошибки
            if ($errcode != 0) {
                return $this->raiseError($errcode);
            }
        } else {
            $this->nak();
        }
        return $ans;
    }
    // полное сообщение в бинарном виде без ведущего STX
    protected function checkCRC($answer) {
        $crc = 0;
        for ($i = 0, $len = strlen($answer); $i < $len; $i++) {
            $byte = $this->byteCode(substr($answer, $i, 1));
            if ($i == ($len - 1)) {
                break;
            }
            $crc ^= $byte;
        }
        return ($crc == $byte);
    }
    protected function raiseError($errorCode) {
        return array(false, 'txt'=>$this->getErrorMessagebyCode($errorCode));
    }
    protected function getSubModeTitleByCode($mode) {
        $const = 'SHTRIH_KKMSUBMODE_' . hexdec($mode);
        if (defined($const)) {
            return constant($const);
        }
        return 'Неизвестный подрежим';
    }
    protected function getModeTitleByCode($mode) {
        $mode = hexdec($mode);
        $submode = ($mode & 0xF0) >> 4;
        $mode = ($mode & 0x0F);
        $const = 'SHTRIH_KKMMODE_' . $mode . '_' . $submode;
        if (defined($const)) {
            return constant($const);
        }
        return 'Неизвестный режим';
    }
    protected function getErrorMessagebyCode($errorCode) {
        $errorconst = 'SHTRIH_ERROR_CODE_' . $errorCode;
        if (defined($errorconst)) {
            return constant($errorconst);
        }
        return 'unknown error';
    }
    // переводит 1000 (5 байт) в e803000000
    protected function prepareArg($num, $bytes) {
        return implode('', array_reverse(str_split(sprintf('%0' . ($bytes * 2) . 'x', $num), 2)));
    }
    protected function prepareText($text, $bytes = 40) {
        if (preg_match('//u', $text)) { // is UTF
            $text = iconv('UTF-8', 'cp1251//IGNORE', $text);
        }
        $text_code = '';
        for ($i = 0, $n = strlen($text); $i < $n; $i++) {
            $byte = sprintf('%x', ord($text{$i}));
            $text_code .= $byte;
        }
        $maxlen = $bytes * 2;
        return str_pad(substr($text_code, 0, $maxlen), $maxlen, '0', STR_PAD_RIGHT); // 40 bytes max!
    }

    public function read_table($table, $row, $cell) {
        $args=$this->prepareArg($table, 1);
        $args.=$this->prepareArg($row, 2);
        $args.=$this->prepareArg($cell, 1);
        $res=$this->writecmd('1F', $args);
        $n=strlen($res);
        $ans='';
        for($i=$n-2; $i>=2; $i-=1) {
            $ans.=unpack('H*', $res{$i})[1];
        }
        return $ans;
    }

    // Операция V2
    public function operation_v2($ticket_type, $price, $qty, $section, $text, $vt, $psr, $ppr) {
        $tax=$this->read_table(6, $vt, 1); // Узнаем ставку НДС в % из таблицы
        $tax=hexdec($tax)/100;
        $tax_c=2**($vt-1); // 2 возведенное в степень равную номеру налоговой группы уменьшенному на еденицу
        $ticket_type = $this->prepareArg($ticket_type, 1);
        $sum=$this->prepareArg($price*$qty, 5);
        $tax_sum=$this->prepareArg(round($price/(100+$tax)*$tax*$qty), 5);
        $tax_c = $this->prepareArg($tax_c, 1);
        $qty = $this->prepareArg($qty*1000000, 6);
        $price = $this->prepareArg($price, 5);
        $vt = $this->prepareArg($vt, 1);
        $section = $this->prepareArg($section, 1);
        $psr = $this->prepareArg($psr, 1);
        $ppr = $this->prepareArg($ppr, 1);
        $args=$ticket_type.$qty.$price.$sum.$tax_sum.$tax_c.$section.$psr.$ppr.$this->prepareText($text, 128);
        return $this->writecmd('ff46', $args);
    }
   
    // Продажа
    // цена передается как дробное. приводится внутри к копейкам
    // $qty передается как дробное, округлется до 3 знаков после запятой
    public function sell($price, $qty, $section, $text = 'sell', $vt=0) {
        $qty = $this->prepareArg($qty * 1000, 5);
        $price = $this->prepareArg(round($price * 100), 5);
        $vt = $this->prepareArg($vt, 1);
        $section = $this->prepareArg($section, 1);
        $args = sprintf('%10s%10s%02d%02d000000%80s', $qty, $price, $section, $vt, $this->prepareText($text, 40));
        return $this->writecmd('80', $args);
    }

    // Возврат продажи
    // цена передается как дробное. приводится внутри к копейкам
    // $qty передается как дробное, округлется до 3 знаков после запятой
    public function return_sell($price, $qty, $section, $text = 'sell', $vt=0) {
        $qty = $this->prepareArg($qty * 1000, 5);
        $price = $this->prepareArg(round($price * 100), 5);
        $vt = $this->prepareArg($vt, 1);
        $section = $this->prepareArg($section, 1);
        $args = sprintf('%10s%10s%02d%02d000000%80s', $qty, $price, $section, $vt, $this->prepareText($text, 40));
        return $this->writecmd('82', $args);
    }


    // сумма - дробная в рублях.
    public function closeCheck($sum, $print_line = '') {
        $args = $this->prepareArg($sum * 100, 5); // сумма наличных
        $args .= $this->prepareArg(0, 5);
        $args .= $this->prepareArg(0, 5);
        $args .= $this->prepareArg(0, 5); // сумма типа оплаты 4
        $args .= $this->prepareArg(0, 2); // скидка / надбавка
        $args .= '00000000'; // налоги
        $args .= $this->prepareText($print_line, 40);
        return $this->writecmd('85', $args);
    }
    public function getCheckSubtotal() {
        $ans = $this->writecmd('89');
        $ans = $this->unpackDigits(substr($ans, 3, 5));
        return ($ans / 100);
    }
    public function getKKMSumm() {
        return $this->getKKMMoneyRegister(241) / 100;
    }
    protected function getKKMMoneyRegister($no) {
        $ans = $this->writecmd('1A', sprintf('%x', $no));
        $ans = $this->unpackDigits(substr($ans, 3, 6));
        return $ans;
    }
    protected function getKKMOperRegister($no) {
        $ans = $this->writecmd('1B', sprintf('%x', $no));
        $ans = $this->unpackDigits(substr($ans, 3, 2));
        return $ans;
    }
    protected function unpackDigits($ans) {
        $ans = unpack('H*', implode('', array_reverse(str_split($ans))));
        $ans = reset($ans);
        return hexdec($ans);
    }
    public function getLastDocumentNumber() {
        return $this->getKKMOperRegister(152);
    }

    public function returnMoney($price, $qty, $section, $simulate = 0) {
    }

    protected function textToHex($text) {
        $text = iconv(CFG_SYSTEM_INTERNAL_ENCODING, 'ibm866//IGNORE', $text);
        $packed = $this->packString($text);
        return $this->unpackBinaryString($packed);
    }

    protected function BCD2num($bcd) {                  // Функция преобразования из BCD в человекочитаемый вид

        for($i=1,$res="";$i<strlen($bcd);++$i) {                // Перебираем побайтно
            $ln = sprintf("%08b",ord($bcd{$i}));                // Получаем строку 8 символов
            $resarr = str_split($ln, 4);                        // Разбиваем на 2 части по 4
            $result .= bindec($resarr[0]).bindec($resarr[1]);   // Переводим в нормальный вид и склеиваем
        }
        return (intval($result))/100;               // Возвращаем float вроде 1234.56
    }

    protected function num2BCD($num) {                  // Функция преобразования в BCD из человекочитаемого вида
        if (floatval(strlen($num)/2*2)!=(strlen($num)/2*2)) $num = "0".$num;

        for($i=0,$res="";$i<strlen($num);$i+=2) {               // Перебираем
            $result .= chr(bindec(decbin($num{$i}).decbin($num{$i+1})));
        }

        return $result;
    }
}
