#!/usr/bin/python3

# -*- coding: UTF-8 -*-
from http.server import BaseHTTPRequestHandler, HTTPServer
from escpos.printer import Usb, Serial, File
from escpos.constants import PAPER_FULL_CUT, PAPER_PART_CUT

import math

class GP(BaseHTTPRequestHandler):
  def do_HEAD(self):
    self.send_response(404)
    self.end_headers()
  def do_GET(self):
    self.send_response(404)
    self.end_headers()

  def do_POST(self):
    length = int(self.headers['content-length'])
    data = self.rfile.read(length) #it is bytes
    data = data.decode("utf-8")

    if self.path == '/req' and data[:5] == 'data=':
      self.send_response(200)
      self.send_header('Content-type', 'text/html')
      self.send_header('Access-Control-Allow-Origin', '*')
      self.end_headers()

      data = data[5:] #strip data=
      data = data.split('&') #remove not used parameters
      data = data[0] #take data parameter only

      resp = self.doPRINT(data) #magic

      self.wfile.write(resp.encode("utf-8")) #bytes(string, "utf-8") or string.encode("utf-8")
    else:
      self.send_response(404)
      self.end_headers()

  def doPRINT(self, data):
    result = '' #Наименование_команды;код_ответа;текст_ответа;ответ_ядра XML BASE64URL;
    connected = True

    kassir = 'Кассир СИСТ.АДМИН'
    check_type = 0 #0 продажа, 1 возврат продажи, 2 покупка, 3 возврат покупки
    check_type_text = {0:"ПРОДАЖА", 1:"ВОЗВРАТ ПРОДАЖИ", 2:"ПОКУПКА", 3:"ВОЗВРАТ ПОКУПКИ"}
    check_text = ''
    check_sum = 0.0
    check_width = 32 #32 #48
    check_width_bold = 32

    oplaty = {0:"Наличными", 1:"Электронно"}

    # Форматируем строку
    def line_format(s, w=32):
      s=str.replace(s, '#kkm_bold#', '');
      # Заменяем #kkm_right# на пробелы
      kkm_right="#kkm_right#" # тэг сдвига вправо
      space_n=w-(len(s)-len(kkm_right)) # кол-во пробелов
      s=str.replace(s, kkm_right, " "*space_n) # заменяем тэг сдвига на пробелы

      # Заменяем #kkm_center# на пробелы
      kkm_center="#kkm_center#" # тэг сдвига к центру
      space_n=math.floor((w-(len(s)-len(kkm_center)))/2) # кол-во пробелов
      s=str.replace(s, kkm_center, " "*space_n) # заменяем тэг сдвига на пробелы

      return s

    try:
      #INIT HERE
      prnt = File("/dev/usb/lp0") #/dev/usb/lp0
      #prnt = File("/dev/ttyS3")
      #prnt = Serial("/dev/ttyS3", 19200)
      #prnt.charcode("CIRILLIC2") #set encoding and send control sequence to printer
      prnt.charcode_nosend("CIRILLIC2") #set encoding NOT send control sequence

    except Exception as e:
      print(e)
      connected = False

    for l in data.strip().splitlines():
      d = l.strip().split(';')

      #reset print settings to default
      if connected:
        prnt.set(align='left', font='a', text_type='normal', width=1, height=1, density=9, invert=False, smooth=False, flip=False)

      if not connected:
        result += d[0]+";255;Нет связи c принтером;;\n"

      elif d[0]=='feed': #feed;КОЛВО_СТРОК; протяжка ленты
        try:
          prnt.text("\n"*int(d[1]))
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";1;"+format(e)+";;\n"

      elif d[0]=='continue_print': #continue_print; Продолжить печать после ошибки
        try:
          prnt.hw("RESET")
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";9;"+format(e)+";;\n"

      elif d[0]=='cut_check': #cut_check; отрезчик
        try:
          #prnt.cut(d[1]) #to many newlines in stock method
          #prnt._raw(PAPER_FULL_CUT)
          prnt._raw(PAPER_PART_CUT)
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";10;"+format(e)+";;\n"

      elif d[0]=='open_cash_box': #open_cash_box;НОМЕР_ЯЩИКА у нас PIN;
        try:
          if int(d[1]) != 2 and int(d[1]) != 5:
            raise ValueError('Корректные значения 2 и 5')
          prnt.cashdraw(int(d[1])) #pin number, 2 or 5
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";11;"+format(e)+";;\n"

      elif d[0]=='p': #p;ТЕКСТ; выводит текст
        try:
          prnt.text(line_format(d[1], check_width)+"\n")
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";2;"+format(e)+";;\n"

      elif d[0]=='pm': #pm;ТЕКСТ; выводит текст c новыми линиями
        try:
          for line in d[1].split("#kkm_br#"):
            if line[0:len('#kkm_bold#')]=='#kkm_bold#':
              line=line_format(line, check_width_bold)
              prnt.set(height=2)
              prnt.text(line+"\n")
              prnt.set(height=1)
            else:
              line=line_format(line, check_width)
              prnt.text(line+"\n")
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";3;"+format(e)+";;\n"

      elif d[0]=='print_font': #pm;НОМЕР_ШРИФТА;ТЕКСТ; выводит текст заданым шрифтом
        try:
          prnt.text(d[2]+"\n")
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";4;"+format(e)+";;\n"

      elif d[0]=='print_bold': #print_bold;ТЕКСТ; выводит текст жирным
        try:
          prnt.set(height=2)
          prnt.text(d[1]+"\n")
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";5;"+format(e)+";;\n"

      elif d[0]=='b': #b;ТИП_ЧЕКА;КАССИР;ПЕЧАТЬ;
        try:
          check_type = int(d[1])
          kassir = d[2]
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";6;"+format(e)+";;\n"

      elif d[0]=='smde': #smde;НАЗВАНИЕ_ТОВАРА;ЦЕНА;КОЛВО;НАЛОГ;СЕКЦИЯ;
        try:
          price = round(float(d[2])/100, 2)
          kolvo = round(float(d[3])/1000, 3)
          summ = round(price*kolvo, 2)

          check_sum += summ

          check_text += d[1]+"\n "
          check_text += '{:02d}'.format(int(d[5])) #" "+d[5]
          #'{:{align}{width}}'.format('text', align='^', width='10')
          check_text += '{:{a}{w}}'.format('{:.3f}'.format(kolvo)+" x "+'{:.2f}'.format(price)+" ="+'{:.2f}'.format(summ), a='>', w=str(check_width-3)) + "\n"

          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";7;"+format(e)+";;\n"

      elif d[0]=='tmde':
        #tmde;СУММА;НОМЕР_ОПЛАТЫ;
        #tmde;СУММА1;СУММА2;СУММА3;СУММА4;НОМЕР_НАЛОГА_1;НОМЕР_НАЛОГА_2;НОМЕР_НАЛОГА_3;НОМЕР_НАЛОГА_4;
        try:
          check_text += "ИТОГО"
          check_text += '{:{a}{w}}'.format("="+'{:.2f}'.format(check_sum), a='>', w=str(check_width-5)) + "\n"

          if 8 < len(d):
            #смешанная оплата, должны быть заполнены все поля
            if int(d[1]) > 0 and 0 < len(oplaty):
              s = round(float(d[1])/100, 2)
              check_text += oplaty[0]
              check_text += '{:{a}{w}}'.format("="+'{:.2f}'.format(s), a='>', w=str(check_width-len(oplaty[0]))) + "\n"

            if int(d[2]) > 0 and 1 < len(oplaty):
              s = round(float(d[2])/100, 2)
              check_text += oplaty[1]
              check_text += '{:{a}{w}}'.format("="+'{:.2f}'.format(s), a='>', w=str(check_width-len(oplaty[1]))) + "\n"

            if int(d[3]) > 0 and 2 < len(oplaty):
              s = round(float(d[3])/100, 2)
              check_text += oplaty[2]
              check_text += '{:{a}{w}}'.format("="+'{:.2f}'.format(s), a='>', w=str(check_width-len(oplaty[2]))) + "\n"

            if int(d[4]) > 0 and 3 < len(oplaty):
              s = round(float(d[4])/100, 2)
              check_text += oplaty[3]
              check_text += '{:{a}{w}}'.format("="+'{:.2f}'.format(s), a='>', w=str(check_width-len(oplaty[3]))) + "\n"
          else:
            #оплата одним способом
            o = int(d[2])
            s = round(float(d[1])/100, 2)

            check_text += oplaty[o]
            check_text += '{:{a}{w}}'.format("="+'{:.2f}'.format(s), a='>', w=str(check_width-len(oplaty[o]))) + "\n"

            #Если закрытие происходит наличными, то СУММА должна быть не менее суммы всех стоимостей продаж по чеку. При превышении -- расчитывается сдача.
            if o == 0: #наличными
              sdacha = round(s-check_sum, 2)
              check_text += "Cдача"
              check_text += '{:{a}{w}}'.format("="+'{:.2f}'.format(sdacha>0 and sdacha or 0), a='>', w=str(check_width-5)) + "\n"

          prnt.text(check_type_text[check_type]+"\n")
          prnt.text(kassir+"\n")
          prnt.text(check_text+"\n")
          result += d[0]+";0;Успешно;;\n"
        except Exception as e:
          result += d[0]+";8;"+format(e)+";;\n"

      else:
        result += d[0]+";0;Команда не поддерживается;;\n"

    return result

def run(server_class=HTTPServer, handler_class=GP, port=8888):
  try:
    server_address = ('0.0.0.0', port)
    httpd = server_class(server_address, handler_class)
    print('Server running at localhost:8888...')
    httpd.serve_forever()
  except KeyboardInterrupt:
    print(' KeyboardInterrupt received, shutting down server')
    httpd.socket.close()

run()
